
# Author: Nguyễn Hồng Quân <ng.hong.quan@gmail.com>
#
# References:
# Legal documents: Circular 36/2010/TT-BCA, Circular 15/2014/TT-BCA
# Internet:
# https://thuvienphapluat.vn/van-ban/Giao-thong-Van-tai/Thong-tu-15-2014-TT-BCA-quy-dinh-ve-dang-ky-xe-Bo-truong-Bo-Cong-an-229224.aspx
# https://danluat.thuvienphapluat.vn/ban-da-biet-gi-ve-bien-so-xe-132978.aspx
# https://tuoitre.vn/2800-xe-bon-banh-khong-biet-xep-vao-loai-xe-gi-1078666.htm

# Note that, in this library, we only use uppercase letters

import re
import enum
from typing import Optional

from memoprop import memoized_property
from pydantic.dataclasses import dataclass

from .utils import split_to_triples


# Use this regex to strip out all illegal characters
REGEX_CLEAN_PLATE_NUMBER = re.compile(r'[^a-zđA-ZĐ0-9]')


class SpecialSeries(enum.Enum):
    DA = 'DA'  # Dự án
    KT = 'KT'  # Quân đội làm kinh tế
    LD = 'LD'  # Liên doanh


class NonBusinessSpecialSeries(enum.Enum):
    MA = 'MA'
    MĐ = 'MĐ'  # Máy điện 29-MĐ1 013.53
    MK = 'MK'  # Máy kéo
    TĐ = 'TĐ'  # Thí điểm 50TĐ-82108
    HC = 'HC'  # Hạn chế phạm vi hoạt động
    SA = 'SA'
    XA = 'XA'
    # Towed vehicles (trailers and semi-trailers)
    # https://banotore.com/xe-xe-tai-tren-10-tan-ba-ria-vung-tau/ban-tra-gop-lai-suat-thap-so-mi-ro-mooc-san-doosung-40feet3truc-aid563816
    R = 'R'   # 51R-14139


# This plate has yellow background
class SpecialEconomicZoneSeries(enum.Enum):
    LB = 'LB'  # Lao Bảo
    CT = 'CT'  # Cầu Treo
    LA = 'LA'  # Don't know where it is
    KL = 'KL'  # Don't know where it is. Found at https://www.dontu.net/2016/10/nhung-ieu-nen-biet-ve-bien-so-xe.html


class VehicleType(enum.IntEnum):
    # Appendix No 4 of Circular 15/2014/TT-BCA
    DOMESTIC_AUTOMOBILE = 1
    # https://soha.vn/xa-hoi/me-man-voi-dan-sieu-xe-bien-cuc-doc-cua-thieu-gia-dat-phu-2013050715292303.htm
    DOMESTIC_MOTORCYCLE_UNDER_50CC = enum.auto()
    DOMESTIC_MOTORCYCLE_50_TO_175CC = enum.auto()
    # https://news.zing.vn/cap-xe-phan-khoi-lon-300-trieu-bi-lam-gia-dang-ky-post631650.html
    DOMESTIC_MOTORCYCLE_OVER_175CC = enum.auto()
    # Automobiles and motorcycles used for the affiliation, project
    # and Military’s vehicles for the purpose of business
    SPECIAL_BUSINESS = enum.auto()
    SPECIAL_ECONOMIC_ZONE = enum.auto()
    # Also tractor and electrical
    NON_BUSINESS_SPECIAL = enum.auto()
    TEMPORARY = enum.auto()
    DIPLOMATIC = enum.auto()
    # Special
    MILITARY = enum.auto()


PATTERN_SPECIAL_SERIES = '|'.join(e.name for e in SpecialSeries)
PATTERN_NON_BUSINESS_SPECIAL_SERIES = '|'.join(e.name for e in NonBusinessSpecialSeries)
PATTERN_SPECIAL_ECONOMIC_ZONE_SERIES = '|'.join(e.name for e in SpecialEconomicZoneSeries)
# Letters used for automobile series registered since 2010
PATTERN_PERMITTED_LETTERS = '[A-HK-NPS-VXYZ]'
# Letters which were used in automobile series before 2010, but no longer used since 2010
PATTERN_OBSOLETE_LETTERS = '[IJOQR]'

REGEXES = {
    # If the plate contains obsolete (pre-2010) letters, the order must be 4-digit.
    # otherwise, the order can be 4 or 5 digits.
    VehicleType.DOMESTIC_AUTOMOBILE: re.compile(
        r'(?P<locality>\d{2})'
        rf'(?P<series>{PATTERN_PERMITTED_LETTERS}|(?P<obsolete_letter>{PATTERN_OBSOLETE_LETTERS}))'
        r'(?P<order>(?(obsolete_letter)\d{4}|\d{4,5}))'
    ),
    # FIXME: This one is quite problematic
    # Example: 80NG-636-70, 80QT-546-42, 80NN-381-35, 80-011-NG-01, 80-631-CV-01
    VehicleType.DIPLOMATIC: re.compile(
        r'(?P<locality>\d{2})'
        r'(?P<country>\d{3})?'
        r'(?P<series>(NG|QT|NN|CV))'
        r'(?P<order>\d{2,5})',
    ),
    VehicleType.SPECIAL_BUSINESS: re.compile(
        r'(?P<locality>\d{2})'
        f'(?P<series>{PATTERN_SPECIAL_SERIES})'
        r'(?P<order>\d{4,5})'
    ),
    VehicleType.SPECIAL_ECONOMIC_ZONE: re.compile(
        r'(?P<locality>\d{2})'
        f'(?P<series>{PATTERN_SPECIAL_ECONOMIC_ZONE_SERIES})'
        r'(?P<order>\d{4,5})'
    ),
    VehicleType.NON_BUSINESS_SPECIAL: re.compile(
        r'(?P<locality>\d{2})'
        f'(?P<series>(?:{PATTERN_NON_BUSINESS_SPECIAL_SERIES})\\d?)'
        r'(?P<order>\d{4,5})'
    ),
    # Be careful that this plate may be mistaken with the special ones above, with two-letter series
    VehicleType.DOMESTIC_MOTORCYCLE_UNDER_50CC: re.compile(
        r'(?P<locality>\d{2})'
        r'(?P<series>[A-HK-NPS-VXYZ][A-FHK-NPR-VXYZ])'
        r'(?P<order>\d{4,5})',
    ),
    VehicleType.DOMESTIC_MOTORCYCLE_50_TO_175CC: re.compile(
        r'(?P<locality>\d{2})'
        r'(?P<series>[B-Z][1-9])'
        r'(?P<order>\d{4,5})'
    ),
    VehicleType.DOMESTIC_MOTORCYCLE_OVER_175CC: re.compile(
        r'(?P<locality>\d{2})'
        r'(?P<series>A[1-9])'
        r'(?P<order>\d{4,5})'
    ),
    # Example: TC3386
    VehicleType.MILITARY: re.compile(r'(?P<series>[A-DHKPQTV][A-X]B?)(?P<order>\d{4})'),
    # The temporary plate doesn't have series
    VehicleType.TEMPORARY: re.compile(r'(?P<series>T)(?P<locality>\d{2})(?P<order>\d{5})'),
}


@dataclass
class VietnamVehiclePlate:
    """Class to represent a Vietnamese vehicle number plate.

    Library user must not create instance directly from this class constructer.
    Please call :meth:`VietnamVehiclePlate.from_string` instead.

    :param vehicle_type: Type of vehicle (of :class:`VehicleType` type), deduced from the plate number.
    :param series: Series string of the plate number.
    :param order: Registration order number.
    :param locality: Locality where this vehicle was registered.
    :param dip_country: Foreign country where the vehicle owner came from (in case of diplomat, foreigner use).
    """

    vehicle_type: VehicleType
    series: str
    order: str
    locality: Optional[str] = None
    dip_country: Optional[str] = None

    @memoized_property
    def compact(self):
        """
        Compact string of plate number, where all characters other than letters and numbers are stripped.

        If we are about to serialize and save this object, the compact form should be used.
        """
        locality = self.locality or ''
        dip_country = self.dip_country or ''
        return f'{locality}{dip_country}{self.series}{self.order}'

    def __len__(self):
        """Return the length of this object, if it is about to be saved some where as string."""
        return len(self.compact)

    def __str__(self):
        """Return string representation of this object."""
        vehicle_type = self.vehicle_type
        mototcycle_types = (VehicleType.DOMESTIC_MOTORCYCLE_UNDER_50CC, VehicleType.DOMESTIC_MOTORCYCLE_OVER_175CC,
                            VehicleType.DOMESTIC_MOTORCYCLE_50_TO_175CC)
        if len(self.order) <= 4:
            order = self.order
        else:
            # For number plate since 2010, the order will be separated with dot,
            # to be easier to read.
            order = '.'.join(split_to_triples(self.order))
        if vehicle_type in mototcycle_types:
            return f'{self.locality}-{self.series} {order}'
        if self.dip_country:
            return f'{self.locality}-{self.dip_country}-{self.series}-{order}'
        return f'{self.locality}{self.series}-{order}'

    @classmethod
    def from_string(cls, number_sequence: str) -> 'VietnamVehiclePlate':
        """Parse the number string of Vietnamese vehicle registration plate.

        :param number_sequence: Number string as printed on the plate.
        :return: :class:`VietnamVehiclePlate` object.
        :raises ValueError: If the number string could not be parsed.
        :raises TypeError: If the passed value is not a string.
        """
        if not isinstance(number_sequence, str):
            raise TypeError(f'Need a string, not {type(number_sequence)} type!')
        compact = REGEX_CLEAN_PLATE_NUMBER.sub('', number_sequence.upper())
        if not compact:
            raise ValueError('Empty string!')
        for vtype, regex in REGEXES.items():
            m = regex.fullmatch(compact)
            if not m:
                continue
            data = {
                'vehicle_type': vtype,
                'order': m.group('order')
            }
            try:
                data['locality'] = m.group('locality')
            except IndexError:
                pass
            try:
                data['series'] = m.group('series')
            except IndexError:
                pass
            try:
                data['dip_country'] = m.group('country')
            except IndexError:
                pass
            return VietnamVehiclePlate(**data)
        else:
            # Not found match
            raise ValueError('Unrecognized plate number')



from typing import Tuple


def split_to_triples(string: str) -> Tuple[str, ...]:
    """Split a string to many substrings of 3-character.

    For example, split "abcdef" to "abc", "def".
    """
    return tuple(string[i: i + 3] for i in range(0, len(string), 3))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['biensoxe', 'biensoxe.django']

package_data = \
{'': ['*']}

install_requires = \
['memoprop>=0.2.0,<0.3.0', 'pydantic>=1.0,<2.0']

setup_kwargs = {
    'name': 'biensoxe',
    'version': '0.8.5',
    'description': 'Library to parse and validate Vietnamese vehicle plate',
    'long_description': "========\nBienSoXe\n========\n\n.. image:: https://badgen.net/pypi/v/biensoxe\n   :target: https://pypi.org/project/biensoxe\n\nLibrary to validate and parse Vietnamese vehicle plate.\n\nThis library is not a computer-vision-based license plate recognition software. It instead is used for validating output of such computer vision software. Imagine that you use camera to track all cars coming in and out off your parking lot, but you don't want to save false result (due to wrong angle of camera, for example), you can use this library to check and remove them.\n\nInstall\n-------\n\n.. code-block:: sh\n\n    pip3 install biensoxe\n\n\nUsage\n-----\n\nCall ``VietnamVehiclePlate.from_string``, passing the number string, to create ``VietnamVehiclePlate`` object.\n\n.. code-block:: python\n\n    >>> from biensoxe import VietnamVehiclePlate\n\n    >>> VietnamVehiclePlate.from_string('44A-112.23')\n    VietnamVehiclePlate(compact='44A11223', vehicle_type=<VehicleType.DOMESTIC_AUTOMOBILE: 1>,\n    series='A', order='11223', locality='44', dip_country=None)\n\n    >>> VietnamVehiclePlate.from_string('41-291-NG-01')\n    VietnamVehiclePlate(vehicle_type=<VehicleType.DIPLOMATIC: 9>, series='NG', order='01', locality='41', dip_country='291')\n\n\nThe method raises ``ValueError`` if the string could not be parsed.\n\nTo format the plate number as in daily life, pass ``VietnamVehiclePlate`` to ``str``:\n\n.. code-block:: python\n\n    >>> plate = VietnamVehiclePlate.from_string('72E101130')\n\n    >>> plate\n    VietnamVehiclePlate(compact='72E101130', vehicle_type=<VehicleType.DOMESTIC_MOTORCYCLE_50_TO_175CC: 3>, series='E1', order='01130', locality='72', dip_country=None)\n\n    >>> str(plate)\n    '72-E1 011.30'\n\nDjango\n~~~~~~\n\nThis library provides a field type, ``VietnamVehiclePlateField``, for Django model. The field will return value as ``VietnamVehiclePlate`` object. Here is example:\n\n.. code-block:: python\n\n    from biensoxe.django import VietnamVehiclePlateField\n\n    def default_plate_number():\n        return VietnamVehiclePlate.from_string('10A 00001')\n\n    class Vehicle(models.Model):\n        plate_number = VietnamVehiclePlateField(max_length=20, default=default_plate_number, unique=True)\n\n    def __str__(self):\n        return str(self.plate_number) or self.pk\n\nNote that this field stores value internally as PostgeSQL ``CIText`` data type, so you can only use this field with PostgreSQL.\nYou also need to activate CITextExtension_ yourself.\n\n\nCredit\n------\n\nBrought to you by `Nguyễn Hồng Quân <author_>`_.\n\n\n.. _CITextExtension: https://docs.djangoproject.com/en/2.2/ref/contrib/postgres/operations/#citextextension\n.. _author: https://quan.hoabinh.vn\n",
    'author': 'Nguyễn Hồng Quân',
    'author_email': 'ng.hong.quan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sunshine-tech/BienSoXe',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Django model field to return VietnamVehiclePlate object."""

from typing import Union, Optional

from django.db.models import Expression
from django.db.backends.postgresql.base import DatabaseWrapper
from django.core.exceptions import ValidationError
from django.contrib.postgres.fields import CICharField
from django.utils.translation import gettext_lazy as _

from ..core import VietnamVehiclePlate


def parse_vehicleplate(number_string: str) -> VietnamVehiclePlate:
    """Validate and parse input string to VietnamVehiclePlate object."""
    try:
        return VietnamVehiclePlate.from_string(number_string)
    except ValueError:
        raise ValidationError(_('Input string does not look like Vietname plate number'))


class VietnamVehiclePlateField(CICharField):
    """Field to store Vietnamese vehicle plate. Stored in PostgreSQL as CIText data type, to enable case-insensitive search.

    Return data as VietnamVehiclePlate type from biensoxe library.
    """

    description = _('Field to store Vietnamese vehicle plate')

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 20)
        super().__init__(*args, **kwargs)

    def from_db_value(self, value: Optional[str],
                      expression: Expression, connection: DatabaseWrapper):
        # Called in all circumstances when the data is loaded from the database,
        # including in aggregates and values() calls.
        if value is None:
            return value
        return parse_vehicleplate(value)

    def to_python(self, value: Union[str, VietnamVehiclePlate, None]):
        # Called by deserialization and during the clean() method used from forms.
        if isinstance(value, VietnamVehiclePlate):
            return value
        if value is None:
            return value
        return parse_vehicleplate(value)

    def get_prep_value(self, value: VietnamVehiclePlate):
        # Convert Python object back to query value
        return value.compact
